PLATFORM_CONFIG_FILE = "platform-config.yml"
PLATFORM_HELPER_VERSION_FILE = ".platform-helper-version"
CODEBASE_PIPELINES_KEY = "codebase_pipelines"
ENVIRONMENTS_KEY = "environments"
DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION = "5"
